/*
 * Copyright (C) 2009 Castrense Savojardo
 * 	Bologna Biocomputing Group
 * 	University of Bologna, Italy
 *  savojard@biocomp.unibo.it
 * 
 * Viterbi.cpp - This file is part of BioCRF
 * 
 * BioCRF is a free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 3 of the License,
 * or (at your option) any later version.
 *
 * BioCRF is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with BioCRF; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 * 
 */
#include "Viterbi.h"
#include "Model.h"
#include "ForwardBackward.h"

namespace algo {

Viterbi::Viterbi(int seq_len, int num_states)
{
	int i;
	this->seq_len = seq_len;
	this->num_states = num_states;
	
	this->beg_idx = 0;
	this->end_idx = 1;
	
	this->scaling = true;
	
	this->scale = new double[seq_len];
	this->memory = new mathlib::DoubleMatrix(num_states, seq_len);
	this->indexes = new int*[num_states];
	for (i=0; i<num_states;i++){
		this->indexes[i] = new int[seq_len];
	}
	
	this->Mi = new mathlib::DoubleMatrix(num_states, num_states);
	this->Vi = new mathlib::DoubleVector(num_states);
}

Viterbi::~Viterbi()
{
	int i;
	if (this->indexes != NULL) {
		for (i=0; i<num_states; i++) {
			if (this->indexes[i] != NULL) {
				delete [] this->indexes[i];
			}
		}
		delete [] this->indexes;
	}
	if (memory != NULL)
		delete memory;
	if (scale != NULL)
		delete scale;
	if (Mi != NULL)
		delete Mi;
	if (Vi != NULL)
		delete Vi;
	
}

void Viterbi::reset(int seq_len)
{
	int i;
	if (this->indexes != NULL) {
		for (i=0; i<num_states; i++) {
			if (this->indexes[i] != NULL) {
				delete [] this->indexes[i];
			}
		}
		delete [] this->indexes;
	}
	delete memory;
	delete scale;
	
	this->seq_len = seq_len;
	this->scale = new double[seq_len];
	this->memory = new mathlib::DoubleMatrix(num_states, seq_len);
	this->indexes = new int*[num_states];
	for (i=0; i<num_states;i++){
		this->indexes[i] = new int[seq_len];
	}
}

void Viterbi::backtrace(data::Sequence & seq, models::Model * model)
{
	int i;
	int beg_lbl, end_lbl;
	
	std::map<int,int>::iterator stlblit;
	
	
	stlblit = model->transducer->stlblmap->find(beg_idx);
	beg_lbl = stlblit->second;
	stlblit = model->transducer->stlblmap->find(end_idx);
	end_lbl = stlblit->second;
	
	seq[seq_len-1].prediction = end_lbl;
	int max_idx = indexes[end_idx][seq_len-1];
	
	for (i=seq_len-2; i>0; i--) {
		stlblit = model->transducer->stlblmap->find(max_idx);
		seq[i].prediction = stlblit->second;
		max_idx = indexes[max_idx][i];
	}
	
	seq[0].prediction = beg_lbl;
	
}

void Viterbi::posterior_viterbi(data::Sequence & seq, ForwardBackward * fb, models::Model * model)
{
	int i, j;
	
	reset(seq.size());
	
	mathlib::DoubleVector * temp = new mathlib::DoubleVector(num_states);
	
	memory->set_column(0, 0.0);
	memory->set(beg_idx, 0, fb->p_i(beg_idx,0));
	
	scale[0] = scaling ? fb->p_i(beg_idx,0) : 1;
	
	for (i=1; i<seq_len; i++) {
		for (j=0; j<num_states; j++) {
			model->transducer->transitions->get_column(temp, j);
			memory->column_multiply(i-1, temp, 1);
			temp->multiply(fb->p_i(j, i));
			int max = temp->max();
			memory->set(j, i, temp->get(max));
			indexes[j][i] = max;
		}
		
		scale[i] = scaling ? memory->sum_column(i) : 1;
		memory->column_multiply(i, 1/scale[i]);
	}
	
	delete temp;
	
	backtrace(seq, model);
}

void Viterbi::viterbi(data::Sequence & seq, models::Model * model, double * weights)
{
	int i, j;
	
	reset(seq.size());
	
	mathlib::DoubleVector * temp = new mathlib::DoubleVector(num_states);
	
	memory->set_column(0, 0.0);
	memory->set(beg_idx, 0, 1.0);
	
	scale[0] = 1.0;
	for (i=0; i<num_states; i++){
		indexes[i][0] = 0;
	}
	
	// computing Mi matrix once
	seq[1].compute_transition_feature_scores(model->featureset, Mi, weights);
	Mi->multiply_entrywise(model->transducer->transitions, 0);
	
	if (! eos) 
		Mi->set_column(end_idx, 1.0);
	if (! bos)
		Mi->set_row(beg_idx, 1.0);
			
	for (i=1; i<seq_len; i++) {
		seq[i].compute_state_feature_scores(model->featureset, seq, i, Vi, weights);
		//seq[i].compute_transition_feature_scores(model->featureset, Mi, weights);
		//Mi->multiply_entrywise(model->transducer->transitions, 0);
		
		for (j=0; j<num_states; j++) {
			Mi->get_column(temp, j);
			memory->column_multiply(i-1, temp, 1);
			int max = temp->max();
			memory->set(j, i, temp->get(max)*Vi->get(j));
			indexes[j][i] = max;
		}
		scale[i] = scaling ? memory->sum_column(i) : 1;
		memory->column_multiply(i, 1/scale[i]);
	}
	
	delete temp;
	backtrace(seq, model);
}

}
