/*
 * Copyright (C) 2009 Castrense Savojardo
 * 	Bologna Biocomputing Group
 * 	University of Bologna, Italy
 *  savojard@biocomp.unibo.it
 * 
 * biocrf.cpp - This file is part of BioCRF
 * 
 * BioCRF is a free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 3 of the License,
 * or (at your option) any later version.
 *
 * BioCRF is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with BioCRF; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 * 
 */
#include <iostream>
#include <fstream>
#include <cstdlib>
#include "ModelFactory.h"
#include "Options.h"
#include "Tokenizer.h"

#define ALL 0
#define TRAIN 1
#define TEST 2
#define PREDICT 3

void
copyright()
{
	std::cout << "****************************************************************************" << std::endl;
	std::cout << "* BioCRF - Grammatical-Restrained CRF tool for Bioinformatics applications *" << std::endl;
	std::cout << "* Copyright (C) 2009 Castrense Savojardo                                   *" << std::endl;
	std::cout << "*   Bologna Biocomputing Group                                             *" << std::endl;
	std::cout << "*   University of Bologna, Italy                                           *" << std::endl;
	std::cout << "*   savojard@biocomp.unibo.it                                              *" << std::endl;
	std::cout << "****************************************************************************" << std::endl;
}

void
usage()
{
	std::cout << "Usage: biocrf <ACTION> <OPTIONS> <FILE>" << std::endl;
	std::cout << std::endl;
	std::cout << "\tACTION  The type of process you want to perform." << std::endl;
	std::cout << "\tOPTIONS The arguments for the ACTION (action-specific)." << std::endl;
	std::cout << "\tFILE    The file on witch the ACTION will be executed (action-specific)." << std::endl;
	std::cout << std::endl;
	std::cout << "ACTION can be one of the following:" << std::endl;
	std::cout << std::endl;
	std::cout << "-all      run both training and testing." << std::endl;
	std::cout << "-train    run training." << std::endl;
	std::cout << "-test     run testing." << std::endl;
	std::cout << "-predict  run predicting unlabelled data." << std::endl;
	std::cout << std::endl;
	std::cout << "All ACTIONs share the following OPTIONS:" << std::endl;
	std::cout << std::endl;
	std::cout << "-m model_file           The model file. For -all and -train ACTIONs this file is WRITTEN and at the end of the procedure model_file contains the trained model. For -test and -predict ACTIONs a trained model_file is needed to compute the test or to predict." << std::endl;
	std::cout << std::endl;
	std::cout << "The following OPTIONS are valid only for -all and -train ACTIONs:" << std::endl;
	std::cout << std::endl;
	std::cout << "-g topology_file        Specify the model topology file." << std::endl;
	std::cout << "-j value                Specify the number of iterations of the training algorithm to be performed." << std::endl;
	std::cout << "-i value                Specify the initial parameters value. Optional, default: 0.0." << std::endl;
	std::cout << "-w value                Specify the half size of the sliding window used to generate state features. Optional, default: 0." << std::endl;
	std::cout << "-s value                Specify the sigma square value used by the training algorithm to govern overfitting. Optional, default: 0.5." << std::endl;
	std::cout << "-e value                Specify the epsilon value for convergence. Optional, default: 0.0001." << std::endl;
	std::cout << std::endl;
	std::cout << "The following OPTIONS are valid only for -all, -test an -predict ACTIONs:" << std::endl;
	std::cout << std::endl;
	std::cout << "-d decoding_algorithm   Decoding algorithm to be used. The string decoding_algorithm can be viterbi or posterior-viterbi." << std::endl;
	std::cout << "-o output_file          Specify the file where predictions will be stored." << std::endl;
	std::cout << "-q posterior_out_file   Store the posterior matrix in a file. This option is used only in case of posterior-viterbi decoding." << std::endl;
	std::cout << std::endl;
	std::cout << "The following OPTIONS are specific of the -all ACTION:" << std::endl;
	std::cout << std::endl;
	std::cout << "-t testing_file         Specify the testing sequences file." << std::endl;
	std::cout << "-p unlabelled_file      Specify the unlabelled sequences file." << std::endl;
	std::cout << std::endl;
	std::cout << "Depends on the ACTION performed, the FILE argument will be considered as follows:" << std::endl;
	std::cout << std::endl;
	std::cout << "-all, -train      FILE is the training sequences file." << std::endl;
	std::cout << "-test             FILE is the testing sequences file." << std::endl;
	std::cout << "-predict          FILE is the unlabelled sequences file." << std::endl;
	std::cout << std::endl;
	std::cout << "Examples:" << std::endl;
	std::cout << std::endl;
	std::cout << "biocrf -all -g topology.model -m model.txt -s 0.005 -j 60 -w 7 -d viterbi -p unlabelled.set -o out.predictions train.set" << std::endl;
	std::cout << "biocrf -train -g topology.model -m model.txt -s 0.005 -j 60 -e 0.0001 -w 7 train.set" << std::endl;
	std::cout << "biocrf -test -m model.txt -d posterior-viterbi -o out.predictions -q posteriors.matrix test.set" << std::endl;
	std::cout << "biocrf -predict -m model.txt -d viterbi -o out.predictions unlabelled.set" << std::endl;
	
}

int
get_opt(int argc, char **argv, options::Options * opt)
{
	int action;
	
	
	if (argc < 7 || argc > 25)
	{
		usage();
		exit(EXIT_FAILURE);
	}
	
	std::string act_s = argv[1];
	
	if (act_s.compare("-all") == 0) {
		action = ALL;
		opt->training_file = argv[argc-1];
		opt->istraining = true;
	} else if (act_s.compare("-train") == 0) {
		action = TRAIN;
		opt->training_file = argv[argc-1];
		opt->istraining = true;
	} else if (act_s.compare("-test") == 0) {
		action = TEST;
		opt->testing_file = argv[argc-1];
		opt->istesting = true;
	} else if (act_s.compare("-predict") == 0) {
		action = PREDICT;
		opt->unlabelled_file = argv[argc-1];
		opt->ispredicting = true;
	} else {
		std::cout << act_s << " is not a valid action." << std::endl;
		usage();
		exit(EXIT_FAILURE);
	}
	
	int i;
	
	bool isp = false;
	bool ist = false;
	bool isg = false;
	bool isj = false;
	bool ism = false;
	bool isd = false;
	bool iso = false;

	
	for (i=2; i<argc-1; i++) {
	  std::string op = argv[i];
	  
	  if (op.compare("-g") == 0) {
	  	if (action != ALL && action != TRAIN) {
	  		std::cout << "Option -g incompatible with action " << act_s << "." << std::endl;
	  		usage();
	  		exit(EXIT_FAILURE);
	  	}
		isg = true;
	  	i++;
	  	if (i < argc)
	  	opt->topology_file = argv[i];
	  	else { 
	  		usage();
	  		exit(EXIT_FAILURE);
	  	}
	  } else if (op.compare("-j") == 0) {
		if (action != ALL && action != TRAIN) {
			std::cout << "Option -j incompatible with action " << act_s << "." << std::endl;
			usage();
			exit(EXIT_FAILURE);
		}

		isj = true;
		i++;
		if (i < argc)
                opt->iterations = atoi(argv[i]);
                else {
			usage();
			exit(EXIT_FAILURE);
		}


	  } else if (op.compare("-i") == 0) {
                if (action != ALL && action != TRAIN) {
                        std::cout << "Option -i incompatible with action " << act_s << "." << std::endl;
                        usage();
                        exit(EXIT_FAILURE);
                }

                i++;
                if (i < argc)
                opt->init_params_value = atof(argv[i]);
                else {
                        usage();
                        exit(EXIT_FAILURE);
                }
	  } else if (op.compare("-m") == 0) {
	  	i++;
		ism = true;
	  	if (i < argc)
	  	opt->model_file = argv[i];
	  	else { 
	  		usage();
	  		exit(EXIT_FAILURE);
	  	}
	  } else if (op.compare("-t") == 0) {
	  	if (action != ALL) {
	  		std::cout << "Option -t incompatible with action " << act_s << "." << std::endl;
	  		usage();
	  		exit(EXIT_FAILURE);
	  	}
	  	
	  	if (isp) {
	  		std::cout << "Option -t incompatible with option -p." << std::endl;
	  		usage();
	  		exit(EXIT_FAILURE);
	  	}
	  	ist = true;
	  	i++;
	  	if (i < argc){
		opt->istesting = true;
	  	opt->testing_file = argv[i];
		} else { 
	  		usage();
	  		exit(EXIT_FAILURE);
	  	}
	  	
	  } else if (op.compare("-s") == 0) {
	  	if (action != ALL && action != TRAIN) {
	  		std::cout << "Option -s incompatible with action " << act_s << "." << std::endl;
	  		usage();
	  		exit(EXIT_FAILURE);
	  	}
	  	
	  	i++;
	  	if (i < argc)
	  	opt->sigma_square = atof(argv[i]);
	  	else { 
	  		usage();
	  		exit(EXIT_FAILURE);
	  	}
	  } else if (op.compare("-e") == 0) {
	  	if (action != ALL && action != TRAIN) {
	  		std::cout << "Option -e incompatible with action " << act_s << "." << std::endl;
	  		usage();
	  		exit(EXIT_FAILURE);
	  	}
	  	
	  	i++;
	  	if (i < argc)
	  	opt->epsilon = atof(argv[i]);
	  	else { 
	  		usage();
	  		exit(EXIT_FAILURE);
	  	}
	  } else if (op.compare("-w") == 0) {
	  	if (action != ALL && action != TRAIN) {
	  		std::cout << "Option -w incompatible with action " << act_s << "." << std::endl;
	  		usage();
	  		exit(EXIT_FAILURE);
	  	}
	  	
	  	i++;
	  	if (i < argc)
	  	opt->window_size = atoi(argv[i]);
	  	else { 
	  		usage();
	  		exit(EXIT_FAILURE);
	  	}
	  } else if (op.compare("-d") == 0) {
	  	if (action != ALL && action != TEST && action != PREDICT) {
	  		std::cout << "Option -d incompatible with action " << act_s << "." << std::endl;
	  		usage();
	  		exit(EXIT_FAILURE);
	  	}
	  	isd = true;
	  	i++;
	  	if (i < argc){
	  		int algo;
	  		std::string ao = argv[i];
	  		if (ao.compare("viterbi") == 0) 
	  			algo = VITERBI;
	  		else if (ao.compare("posterior-viterbi") == 0)
	  			algo = POSTERIORSUM;
	  		else {
	  			std::cout << "Invalid decoding algorithm " << ao << "." << std::endl;
	  			usage();
	  			exit(EXIT_FAILURE);
	  		}
	  		opt->decoding_algorithm = algo;
	  	} else { 
	  		usage();
	  		exit(EXIT_FAILURE);
	  	}
	  } else if (op.compare("-q") == 0) {
		if (action != ALL && action != TEST && action != PREDICT) {
		        std::cout << "Option -q incompatible with action " << act_s << "." << std::endl;
		        usage();
		        exit(EXIT_FAILURE);
		}
		i++;
                if (i < argc){
			opt->out_posterior = true;
			opt->posterior_out_file = argv[i];
		} else {
                        usage();
                        exit(EXIT_FAILURE);
                }

	  } else if (op.compare("-p") == 0) {
	  	if (action != ALL) {
	  		std::cout << "Option -p incompatible with action " << act_s << "." << std::endl;
	  		usage();
	  		exit(EXIT_FAILURE);
	  	}
	  	
	  	if (ist) {
	  		std::cout << "Option -p incompatible with option -t." << std::endl;
	  		usage();
	  		exit(EXIT_FAILURE);
	  	}
	  	isp = true;
	  	i++;
	  	if (i < argc){
	  	opt->ispredicting = true;
		opt->unlabelled_file = argv[i];
		}
	  } else if (op.compare("-o") == 0) {
	  	if (action != ALL && action != TEST && action != PREDICT) {
	  		std::cout << "Option -o incompatible with action " << act_s << "." << std::endl;
	  		usage();
	  		exit(EXIT_FAILURE);
	  	}
	  	iso = true;
	  	i++;
	  	if (i < argc){
	  		opt->prediction_file = argv[i];
	  	} else { 
	  		usage();
	  		exit(EXIT_FAILURE);
	  	}
	  }
	}

	if (! ism) 
	{
		std::cout << "Please specify the model file with -m options." << std::endl;
		usage();
                exit(EXIT_FAILURE);
	}
	
	if (action == ALL)
	{
		if (!isp && !ist)
		{
			std::cout << "Neither -p nor -t options specified with action -all." << std::endl;
	                usage();
	                exit(EXIT_FAILURE);
		}
	}

	if (action == ALL || action == TRAIN)
	{
		if (! isg) 
		{
			std::cout << "Option -g not specified with action " << act_s << "." << std::endl;
			usage();
			exit(EXIT_FAILURE);
		}
		
                if (! isj)
                {
                        std::cout << "Option -j not specified with action " << act_s << "." << std::endl;
                        usage();
                        exit(EXIT_FAILURE);
                }


	}

	if (action == ALL || action == TEST || action == PREDICT)
	{
		if (! iso) 
		{
			std::cout << "Option -o not specified with action " << act_s << "." << std::endl;
                        usage();
                        exit(EXIT_FAILURE);
		}

		if (! isd)
		{
			std::cout << "Option -d not specified with action " << act_s << "." << std::endl;
                        usage();
                        exit(EXIT_FAILURE);
		}

	}

	return action;
}

int
main(int argc, char **argv) {
	
	options::Options * opt = new options::Options;;
	
	copyright();
	
	int action = get_opt(argc, argv, opt);
	
	models::ModelFactory * mf = new models::ModelFactory;
	models::Model * model;
	
	if (action == ALL) 
	{
		bool oldpost = opt->out_posterior;
		bool oldt = opt->istesting;
		bool oldp = opt->ispredicting;
		opt->out_posterior = false;
		opt->istesting = false;
		opt->ispredicting = false;
		models::Model * model = mf->get_model(opt);
                model->train();

		model->dump();

                delete model;

		opt->out_posterior = oldpost;
		opt->istraining = false;
		opt->istesting = oldt;
		opt->ispredicting = oldp;

		model = mf->get_model(opt);
                model->decode();
                model->data->write_predictions();
	}
	else if (action == TRAIN) 
	{
		model = mf->get_model(opt);
                model->train();
                model->dump();
	}
	else if (action == TEST) 
	{
		model = mf->get_model(opt);
		model->decode();
                model->data->write_predictions();

	}
	else if (action == PREDICT)
	{
		model = mf->get_model(opt);
                model->decode();
                model->data->write_predictions();
	}
	
	exit(EXIT_SUCCESS);
}
