/*
 * Copyright (C) 2009 Castrense Savojardo
 * 	Bologna Biocomputing Group
 * 	University of Bologna, Italy
 *  savojard@biocomp.unibo.it
 * 
 * DoubleMatrix.cpp - This file is part of BioCRF
 * 
 * BioCRF is a free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 3 of the License,
 * or (at your option) any later version.
 *
 * BioCRF is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with BioCRF; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 * 
 */
#include <iostream>
#include <stdio.h>
#include "DoubleMatrix.h"
#include "MathlibException.h"


namespace mathlib
{

DoubleMatrix::DoubleMatrix(int rows, int cols)
{
	try {
	
	if (rows > 0 && cols > 0) {
	this->rows = rows;
	this->cols = cols;
	
	//mtrx = gsl_matrix_alloc(rows, cols);
	mtrx = new double*[rows];
	for (int i=0; i<rows; i++){
		mtrx[i] = new double[cols];
	}
	
	} else {
		throw MathlibNoValidSizeException("DoubleMatrix::DoubleMatrix");
	}
	} catch (MathlibException & e) {
		e.report();
	}
	 
}

void DoubleMatrix::operator=(double val)
{
	int i, j;
	
	for (i=0; i<rows; i++)
		for (j=0; j<cols; j++)
			mtrx[i][j] = val;
}

void DoubleMatrix::operator=(DoubleMatrix & m)
{
	int i, j;
	try {
	if (m.rows == rows && m.cols == cols) {
	for (i=0; i<rows; i++)
		for (j=0; j<cols; j++)
			mtrx[i][j] = m.get(i,j);
	} else {
		throw MathlibIncompatibleSizeException("DoubleMatrix::=");
	}
	} catch (MathlibException & e) {
		e.report();
	}
}

double DoubleMatrix::get(int i, int j)
{
	double res;
	try {
	if (i >= 0 && i < rows && j >= 0 && j < cols) {
	res = mtrx[i][j];
	} else {
		throw MathlibIndexOutOfRangeException("DoubleMatrix::get");
	}
	} catch (MathlibException & e) {
		e.report();
	}
	return res;
}

void DoubleMatrix::set(int i, int j, double value)
{
	try {
	if (i >= 0 && i < rows && j >= 0 && j < cols) {
	//gsl_matrix_set(mtrx, i, j, value);
	mtrx[i][j] = value;
	} else {
		throw MathlibIndexOutOfRangeException("DoubleMatrix::set");
	}
	} catch (MathlibException & e) {
		e.report();
	}
	
}

void DoubleMatrix::set_column(int column, double value)
{
	try {
		if (column >= 0 && column < cols) {
			for (int j=0; j<rows; j++) {
				//gsl_matrix_set(mtrx, j, column, value);
				mtrx[j][column] = value;
			}
		} else {
			throw MathlibIndexOutOfRangeException("DoubleMatrix::set_column(int, double)");
		}
	} catch (MathlibException & e) {
		e.report();
	}
}

void DoubleMatrix::set_column(int column, DoubleVector * v)
{
	try {
		if (column >= 0 && column < cols) {
			if (v->size == rows) {
			for (int j=0; j<rows; j++) {
				//gsl_matrix_set(mtrx, j, column, v->get(j));
				mtrx[j][column] = v->get(j);
			}
			} else {
				throw MathlibIncompatibleSizeException("DoubleMatrix::set_column(int, DoubleVector *)");
			}
		} else {
			throw MathlibIndexOutOfRangeException("DoubleMatrix::set_column(int, DoubleVector *)");
		}
	} catch (MathlibException & e) {
		e.report();
	}
}

void DoubleMatrix::set_row(int row, DoubleVector * v)
{
	try {
		if (row >= 0 && row < rows) {
			if (v->size == cols) {
			for (int j=0; j<cols; j++) {
				//gsl_matrix_set(mtrx, j, row, v->get(j));
				mtrx[row][j] = v->get(j);
			}
			} else {
				throw MathlibIncompatibleSizeException("DoubleMatrix::set_row(int, DoubleVector *)");
			}
		} else {
			throw MathlibIndexOutOfRangeException("DoubleMatrix::set_row(int, DoubleVector *)");
		}
	} catch (MathlibException & e) {
		e.report();
	}
}

void DoubleMatrix::set_row(int row, double value)
{
	try {
		if (row >= 0 && row < rows) {
			for (int j=0; j<cols; j++) {
				mtrx[row][j] = value;
			}
		} else {
			throw MathlibIndexOutOfRangeException("DoubleMatrix::set_row(int, double)");
		}
	} catch (MathlibException & e) {
		e.report();
	}
}

void DoubleMatrix::multiply_entrywise(DoubleMatrix * m, int direction)
{
	int i;
	DoubleVector temp(this->rows);
	
	try {
	if (m->rows == rows && m->cols == cols) {
	for (i=0; i<this->cols; i++) {
		get_column(&temp, i);
		m->column_multiply(i, &temp, !direction);
		if (!direction) {
			set_column(i, &temp);		
		}
	}
	} else {
		throw MathlibIncompatibleSizeException("DoubleMatrix::multiply_entrywise");
	}
	} catch (MathlibException & e) {
		e.report();
	}
}

void DoubleMatrix::multiply(DoubleMatrix * result, int column, DoubleVector * v, 
								int transposition)
{	
	int i;
	
	
	DoubleVector temp(result->rows);
	try {
				

	if (! transposition){
		if (result->rows == rows && v->size == cols) {
			for (i=0; i<this->rows; i++) {
				this->get_row(&temp, i);
				temp.multiply(v);
				result->set(i, column, temp.sum());
			}
		} else {
			throw MathlibIncompatibleSizeException("DoubleMatrix::multiply(DoubleMatrix *, int, DoubleVector *, int)");
		}
	} else {
		if (result->rows == this->cols && v->size == this->rows) {
			for (i=0; i<this->cols; i++) {
				this->get_column(&temp, i);
				temp.multiply(v);
				result->set(i, column, temp.sum());		
			}
		} else {
			throw MathlibIncompatibleSizeException("DoubleMatrix::multiply(DoubleMatrix *, int, DoubleVector *, int)");
		}
	}
	
	} catch (MathlibException & e) {
		e.report();
	}
	
	
	
}

void DoubleMatrix::multiply(DoubleVector * result, DoubleVector * v, 
								int transposition)
{	
	int i;
	DoubleVector temp(result->size);
	
	try {
	
	if (! transposition){
	if (result->size == rows && v->size == cols) {
	for (i=0; i<rows; i++) {
		this->get_row(&temp, i);
		temp.multiply(v);
		result->set(i, temp.sum());
	}
	} else {
		throw MathlibIncompatibleSizeException("DoubleMatrix::multiply(DoubleVector *, DoubleVector *, int)");
	}
	} else {
	if (result->size == cols && v->size == rows) {
	for (i=0; i<cols; i++) {
		this->get_column(&temp, i);
		temp.multiply(v);
		result->set(i, temp.sum());		
	}
	} else {
		throw MathlibIncompatibleSizeException("DoubleMatrix::multiply(DoubleVector *, DoubleVector *, int)");
	}
	}
	} catch (MathlibException & e) {
		e.report();
	}
}

void DoubleMatrix::column_multiply(int column, DoubleVector * v, int direction)
{
	DoubleVector temp(rows);
	
	try {
	if (v->size	== rows) {
	
	get_column(&temp, column);
	
	if (direction) {
	
		v->multiply(&temp);
	} else {
		temp.multiply(v);
		set_column(column, &temp);
	}
	} else {
		throw MathlibIncompatibleSizeException("DoubleMatrix::column_multiply(int, DoubleVector *, int)");
	}
	} catch (MathlibException & e) {
		e.report();
	} 
	
}

void DoubleMatrix::column_multiply(int column, double value)
{
	DoubleVector temp(rows);
	try {
	if (column >= 0 && column < cols) {
	get_column(&temp, column);
	
	temp.multiply(value);
	
	set_column(column, &temp);
	} else {
		throw MathlibIndexOutOfRangeException("DoubleMatrix::column_multiply(int, double)");
	}
	} catch (MathlibException & e) {
		e.report();
	} 
}

void DoubleMatrix::row_multiply(int row, DoubleVector * v, int direction)
{
	DoubleVector temp(cols);
	
	get_row(&temp, row);
	
	if (direction) {
	
		v->multiply(&temp);
	} else {
		temp.multiply(v);
		set_row(row, &temp);
	}
}

void DoubleMatrix::get_row(DoubleVector * row_vector, int row) {
	int j;
	try {
	if (row < rows && row >= 0) { 
	for (j=0; j<cols; j++)
		//row_vector->set(j, gsl_matrix_get(mtrx, row, j));
		row_vector->set(j, mtrx[row][j]);
	} else {
		throw MathlibIndexOutOfRangeException("DoubleMatrix::get_row");
	}
	} catch (MathlibIndexOutOfRangeException & e) {
		e.report();
	}
}

void DoubleMatrix::get_column(DoubleVector * column_vector, int column) {
	int j;
	try {
	if (column < cols && column >= 0) {
	for (j=0; j<rows; j++)
		//column_vector->set(j, gsl_matrix_get(mtrx, j, column));
		column_vector->set(j, mtrx[j][column]);
	} else {
		throw MathlibIndexOutOfRangeException("DoubleMatrix::get_column");
	}
	} catch (MathlibIndexOutOfRangeException & e) {
		e.report();
	}
}

double DoubleMatrix::sum_row(int row) {
	int j;
	double res = 0.0;
	try {
	if (row < rows && row >= 0) { 
	for (j=0; j<cols; j++)
		//res += gsl_matrix_get(mtrx, row, j);
		res += mtrx[row][j];
	} else {
		throw MathlibIndexOutOfRangeException("DoubleMatrix::sum_row");
	}
	} catch (MathlibIndexOutOfRangeException & e) {
		e.report();
	}
	return res;	
}

double DoubleMatrix::sum_column(int column) {
	int j;
	double res = 0.0;
	try {
	if (column < cols && column >= 0) {
	for (j=0; j<rows; j++)
		//res += gsl_matrix_get(mtrx, j, column);
		res += mtrx[j][column];
	} else {
		throw MathlibIndexOutOfRangeException("DoubleMatrix::sum_column");
	}
	} catch (MathlibIndexOutOfRangeException & e) {
		e.report();
	}
	return res;	
}

void DoubleMatrix::sum_column(int column, DoubleVector * vect)
{
	int j;
	
	try {
	if (column < cols && column >= 0 && vect->size == rows) {
	for (j=0; j<rows; j++)
		//res += gsl_matrix_get(mtrx, j, column);
		mtrx[j][column] += vect->get(j);
	} else {
		throw MathlibIndexOutOfRangeException("DoubleMatrix::sum_column");
	}
	} catch (MathlibIndexOutOfRangeException & e) {
		e.report();
	}
	
}

void DoubleMatrix::dump() {
	int i, j;
	
	for (i=0; i<rows; i++){
		for (j=0; j<cols; j++) {
		  //std::cout << this->get(i, j) << " ";
		  printf("%.6f ", this->get(i, j));
		}
		std::cout << std::endl;
	}
}

void DoubleMatrix::dump_column(int j) 
{
	int i;
	for (i=0; i<rows; i++) {
	 printf("%.4f ", this->get(i, j));
	}
	std::cout << std::endl;
}
}

