/*
 * Copyright (C) 2009 Castrense Savojardo
 * 	Bologna Biocomputing Group
 * 	University of Bologna, Italy
 *  savojard@biocomp.unibo.it
 * 
 * Transducer.cpp - This file is part of BioCRF
 * 
 * BioCRF is a free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 3 of the License,
 * or (at your option) any later version.
 *
 * BioCRF is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with BioCRF; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 * 
 */
#include <stdlib.h>
#include <iostream>
#include <fstream>
#include "Transducer.h"
#include "DoubleMatrix.h"
#include "Tokenizer.h"
#include "GrammarException.h"
#include "Data.h"
#include "Model.h"

namespace grammar {

Transducer::Transducer(std::ifstream * model_file)
{
	char buffer[BUFF_SIZE_LONG];
	
	model_file->getline(buffer, BUFF_SIZE_LONG);
	read_topology(model_file);
	model_file->getline(buffer, BUFF_SIZE_LONG);

	this->bw = NULL;
	this->fw = NULL;
}

Transducer::Transducer(std::string topology_file)
{
	/*
	std::filebuf fb;
	fb.open(topology_file.c_str(), std::ios::in);
	std::istream is(&fb);
	*/
	std::ifstream is(topology_file.c_str());
	if (! is.is_open())
	{
		std::cout << "Error opening topology file" << std::endl;
		exit(EXIT_FAILURE);
	}
	read_topology(&is);
	
	is.close();
	
	this->bw = NULL;
	this->fw = NULL;
}

Transducer::~Transducer()
{	
	if (transitions != NULL)
		delete transitions;
	if (bw != NULL)
		delete bw;
	if (labelsmap != NULL)
		delete labelsmap;
	if (dictstint != NULL)
		delete dictstint;
	if (dictststr != NULL)
		delete dictststr;
	if (stlblmap != NULL)
		delete stlblmap;
	if (lblstmmap != NULL)
		delete lblstmmap;
	if (dictlbint != NULL)
		delete dictlbint;
	if (dictlbstr != NULL)
		delete dictlbstr;
}

void Transducer::read_topology(std::ifstream * is)
{
	utils::Tokenizer * tok;
	std::string line;
	char buffer[BUFF_SIZE_LONG];
	
	
	is->getline(buffer, BUFF_SIZE_LONG);
	line = buffer;
	tok = new utils::Tokenizer(line, " \t\r\n");
	
	try {
	
	if (tok->token(0).compare("labels") != 0) {
		throw GrammarLabelsException(); 
	}
	
	num_labels = atoi(tok->token(1).c_str());
	
	if (num_labels < 2)
		throw GrammarLabelsException();
	
	} catch (GrammarException & e) {
		e.report();
	}
	
	is->getline(buffer, BUFF_SIZE_LONG);
	line = buffer;
	
	delete tok;
	tok = new utils::Tokenizer(line, " \t\r\n");
	
	
	try {
	
	if (tok->token(0).compare("states") != 0) {
		throw GrammarNumberOfStatesException(); 
	}
	
	num_states = atoi(tok->token(1).c_str());
	
	if (num_states < 2)
		throw GrammarNumberOfStatesException();
	
	} catch (GrammarException & e) {
		e.report();
	}

	num_labels += 2;
	num_states += 2;
	
	labelsmap = new mathlib::DoubleMatrix(num_states, num_labels);
	*labelsmap = 0.0;
	
	transitions = new mathlib::DoubleMatrix(num_states, num_states);
	*transitions = 0.0;
	
	dictstint = new ist2str;
	dictststr = new sst2int;
	dictlbint = new ilb2str;
	dictlbstr = new slb2int;
	stlblmap = new ist2lbl;
	lblstmmap = new ilb2sta;
	

	dictlbint->insert(std::pair<int,std::string>(0,"BEGIN_L"));
        dictlbstr->insert(std::pair<std::string,int>("BEGIN_L",0));
	dictlbint->insert(std::pair<int,std::string>(1,"END_L"));
        dictlbstr->insert(std::pair<std::string,int>("END_L",1));
	

	dictstint->insert(std::pair<int,std::string>(0,"BEGIN"));
        dictststr->insert(std::pair<std::string,int>("BEGIN",0));

	stlblmap->insert(std::pair<int,int>(0,0));
        lblstmmap->insert(std::pair<int,int>(0,0));

	dictstint->insert(std::pair<int,std::string>(1,"END"));
        dictststr->insert(std::pair<std::string,int>("END",1));
		        
        stlblmap->insert(std::pair<int,int>(1,1));
        lblstmmap->insert(std::pair<int,int>(1,1));

        labelsmap->set(0, 0, 1.0);
	labelsmap->set(1, 1, 1.0);




	is->getline(buffer, BUFF_SIZE_LONG);
	line = buffer;
	delete tok;
	tok = new utils::Tokenizer(line, " \t\r\n");
	
	try {
	if (tok->token(0).compare("label_mapping") != 0) {
		throw GrammarLabelMappingSectionException();
	}
	} catch (GrammarException & e) {
		e.report();
	}
	
	is->getline(buffer, BUFF_SIZE_LONG);
	line = buffer;
	delete tok;
	tok = new utils::Tokenizer(line, " \t\r\n");
	
	while (tok->token(0).compare("end_label_mapping") != 0) {
		
		int lbl;
		std::string lblstr;
		
		try {
		if (tok->count_tokens() != 1) {
			throw GrammarLabelMappingSectionException();
		}
		
		lbl = dictlbint->size();//atoi(tok->token(0).c_str());
		lblstr = tok->token(0);//tok->token(1);
		
		if (lbl < 0 || lbl >= num_labels)
			throw GrammarLabelMappingSectionException();
		
		} catch (GrammarException & e) {
			e.report();
		}

		dictlbint->insert(std::pair<int,std::string>(lbl,lblstr));
		dictlbstr->insert(std::pair<std::string,int>(lblstr,lbl));
		
		is->getline(buffer, BUFF_SIZE_LONG);
		line = buffer;
		delete tok;
		tok = new utils::Tokenizer(line, " \t\r\n");
	}
	
	is->getline(buffer, BUFF_SIZE_LONG);
	line = buffer;
	delete tok;
	tok = new utils::Tokenizer(line, " \t\r\n");
	
	try {
	if (tok->token(0).compare("state_mapping") != 0) {
		throw GrammarStateMappingSectionException();
	}
	} catch (GrammarException & e) {
		e.report();
	}
	
	is->getline(buffer, BUFF_SIZE_LONG);
	line = buffer;
	delete tok;
	tok = new utils::Tokenizer(line, " \t\r\n");
	
	while (tok->token(0).compare("end_state_mapping") != 0) {
		
		int st;//, label;
		std::string ststr;
		std::map<std::string,int>::iterator lmap;
		
		try {
		if (tok->count_tokens() != 2) {
			throw GrammarStateMappingSectionException();
		}
		
		st = dictstint->size();//atoi(tok->token(0).c_str());
		ststr = tok->token(0);//tok->token(1);
		lmap = dictlbstr->find(tok->token(1));//atoi(tok->token(2).c_str());
			
		if (st < 0 || st >= num_states)
			throw GrammarStateMappingSectionException();
		if (lmap == dictlbstr->end())
			throw GrammarLabelsException();
		
		} catch (GrammarException & e) {
			e.report();
		}

		dictstint->insert(std::pair<int,std::string>(st,ststr));
		dictststr->insert(std::pair<std::string,int>(ststr,st));
		
		stlblmap->insert(std::pair<int,int>(st,lmap->second));
		lblstmmap->insert(std::pair<int,int>(lmap->second,st));
		
		labelsmap->set(st, lmap->second, 1.0);
		
		is->getline(buffer, BUFF_SIZE_LONG);
		line = buffer;
		delete tok;
		tok = new utils::Tokenizer(line, " \t\r\n");
	}
	
	is->getline(buffer, BUFF_SIZE_LONG);
	line = buffer;
	delete tok;
	tok = new utils::Tokenizer(line, " \t\r\n");
	
	
	try {
	if (tok->token(0).compare("transitions") != 0) {
		throw GrammarStateTransitionsSectionException();
	}
	} catch (GrammarException & e) {
		 e.report();
	}
	
	/*
	int current_pos = is->tellg();
	
	is->seekg(0,std::ios::end);
	int flen = is->tellg();
	is->clear();
	is->seekg(flen - 16);
	
	is->getline(buffer, BUFF_SIZE_LONG);
	line = buffer;
	delete tok;
	tok = new utils::Tokenizer(line, " \t\r\n");
	
	try {
	if (tok->token(0).compare("end_transitions") != 0) {
		throw GrammarStateTransitionsSectionException();
	}
	} catch (GrammarException & e) {
		 e.report();
	}
	
	is->clear();
	is->seekg(current_pos);
	*/
	is->getline(buffer, BUFF_SIZE_LONG);
	line = buffer;
	delete tok;
	tok = new utils::Tokenizer(line, " \t\r\n");
	
	while (tok->token(0).compare("end_transitions") != 0) {
		int len = tok->count_tokens();

		std::map<std::string,int>::iterator smap;
		smap = dictststr->find(tok->token(0));
		
		//int from;//atoi(tok->token(0).c_str());
		try {

		if (smap == dictststr->end()) {
			throw GrammarStateTransitionsSectionException();
		}
		
		int from = smap->second;

		for (int k = 1; k<len;k++) {
			
			std::map<std::string,int>::iterator tosmap = dictststr->find(tok->token(k));

			if (tosmap == dictststr->end()) {
				throw GrammarStateTransitionsSectionException();
			}
			
			int to = tosmap->second;//atoi(tok->token(k).c_str());

			transitions->set(from, to, 1.0);
		}
		} catch (GrammarException & e) {
		 e.report();
		}
		is->getline(buffer, BUFF_SIZE_LONG);
		line = buffer;
		delete tok;
		tok = new utils::Tokenizer(line, " \t\r\n");
	}
	
	delete tok;
}

void Transducer::simulate(data::Sequence & seq) 
{
	int i, j;
	
	int seq_len = seq.size();
	
	delete bw;
	
	fw = new mathlib::DoubleMatrix(num_states, seq_len);
	bw = new mathlib::DoubleMatrix(num_states, seq_len);
	
	int beg_idx = dictststr->find("BEGIN")->second;
	int end_idx = dictststr->find("END")->second;
	
	bw->set_column(seq_len-1, 0.0);
	bw->set(end_idx, seq_len-1, 1.0);
	mathlib::DoubleVector temp(num_states);
	for (i=seq_len-2; i>=0; i--) {
		bw->get_column(&temp, i+1);
		transitions->multiply(bw, i, &temp, 0);
		labelsmap->get_column(&temp, seq[i].label);
		bw->column_multiply(i, &temp, 0);
	}
	
	fw->set_column(0, 0.0);
	fw->set(beg_idx, 0, 1.0);
	
	fw->get_column(&temp, 0);
	bw->column_multiply(0, &temp, 0);
	
	for (i=1; i<seq_len; i++) {
		fw->get_column(&temp, i-1);
		transitions->multiply(fw, i, &temp, 1);
		labelsmap->get_column(&temp, seq[i].label);
		fw->column_multiply(i, &temp, 0);
		fw->get_column(&temp, i);
		bw->column_multiply(i, &temp, 0);
	}
	if (bw->get(beg_idx, 0) == 0.0){
		std::cout << "Sequence  does not agree with grammar rules. Exiting." << std::endl;
		for (i=0; i<seq_len; i++) {
			std::cout << seq[i].label << " ";
			for (int k=0; k<num_states; k++){
			  if (bw->get(k,i) > 0){
			  	std::cout << k << " ";
			  }
			  if (fw->get(k,i) > 0){
			       std::cout << k << " ";
			  }

			}
			std::cout << std::endl;
		}
		exit(EXIT_FAILURE);
	}
	/*

	mathlib::DoubleVector * temp = new mathlib::DoubleVector(num_states);
	mathlib::DoubleVector * temp2 = new mathlib::DoubleVector(num_states);
	
	std::map<std::string,int>::iterator mit = dictststr->find("BEGIN");
	*fw = 0.0;
	fw->set(mit->second, 0, 1.0);
	mit = dictststr->find("END");
	*bw = 0.0;
	bw->set(mit->second, len-1, 1.0);
	
	
	int threshold;
	if ((len % 2) == 1)
		threshold = (len-1)/2;
	else
		threshold = (len+1)/2;
	
	for (j = 1; j < len; j++) {
		
		int curr_fw = j;
		int prev_fw = j - 1;
		int curr_bw = len - j - 1;
		int next_bw = len - j;
		
		for (int k = 0; k < num_states; k++) {
			
			std::map<int,int>::iterator lmit = stlblmap->find(k);
			double val;
			
			if (lmit->second == seq[curr_fw].label) {
				transitions->get_column(temp, k);
				fw->column_multiply(prev_fw, temp, 1);
				val = temp->sum();
				if (val>1) val=1;
			} else {
				val = 0.0;
			}
			fw->set(k, curr_fw, val);
			
			if (lmit->second == seq[curr_bw].label) {
				transitions->get_row(temp, k);
				bw->column_multiply(next_bw, temp, 1);
				val = temp->sum();
				if (val>1) val=1;
			} else {
				val = 0.0;
			}
			bw->set(k, curr_bw, val);
		}
		
		if (curr_fw > threshold ) {
			fw->get_column(temp, curr_fw-1);
			fw->get_column(temp2, curr_bw+1);
			
			bw->column_multiply(curr_fw-1, temp, 1);
			bw->column_multiply(curr_bw+1, temp2, 1);
			
			fw->set_column(curr_fw-1, temp);
			fw->set_column(curr_bw+1, temp2);
		} 
	}
	
	bw->get_column(temp, 0);
	fw->column_multiply(0, temp, 0);
	
	bw->get_column(temp, len-1);
	fw->column_multiply(len-1, temp, 0);
	
	delete temp;
	delete temp2;
	*/
	
	
	
	
	
	
	delete fw;
}

std::vector<int> Transducer::get_states(int pos)
{
	int k;
	std::vector<int> res;
	for (k=0; k<num_states; k++) {
		if (bw->get(k, pos) > 0.0) {
			res.push_back(k);
		}
	}
	/*
	if (pos == bw->cols-1)
		delete bw;
		*/
	return res;
}

void Transducer::dump(std::ofstream * os)
{
	/*
	 * std::filebuf fb;
	fb.open(this->model->options->topology_file.c_str(), std::ios::in);
	std::istream is(&fb);
	*/
	std::ifstream is(this->model->options->topology_file.c_str());
	if (! is.is_open()) {
                std::cout << "Error opening topology file" << std::endl;
                exit(EXIT_FAILURE);
       		      
	}
	char buffer[BUFF_SIZE_LONG];
	std::string line;
	
	*os << "#TOPOLOGY" << std::endl;
	
	while (! is.getline(buffer, BUFF_SIZE_LONG).eof()) {
		line = buffer;
		*os << line << std::endl;
	}
	
	*os << "#END_TOPOLOGY" << std::endl;
	is.close();
}

}
